/*
 * RGB-Matrix-Uhr 64x32, Scott-Falk Hhn, Matrix-Uhr_asm.s
 * AtmelStudio Editor-Einstellung: Tools / Options / Text Editor / Assembler / Tabs / Block, Tab size: 8, Indent size: 8, Insert spaces
 * Einige Ideen und Tricks stammen aus der RGBmatrixPanel-Library fr Arduino von Adafruit, vielen Dank an die Entwickler.
 *
 * Die LED-Matrix verwendet folgende I/O-Ports:
 *
 * PB0 - LED-Matrix - Ausgang CLK
 * PA0 - LED-Matrix - Ausgang A
 * PA1 - LED-Matrix - Ausgang B
 * PA2 - LED-Matrix - Ausgang C
 * PA3 - LED-Matrix - Ausgang D, bei Matrix-Scan-Modus 1/8 nicht genutzt
 * PA4 - LED-Matrix - Ausgang LAT
 * PA5 - LED-Matrix - Ausgang OE
 * PC2 - LED-Matrix - Ausgang R1
 * PC3 - LED-Matrix - Ausgang G1
 * PC4 - LED-Matrix - Ausgang B1
 * PC5 - LED-Matrix - Ausgang R2
 * PC6 - LED-Matrix - Ausgang G2
 * PC7 - LED-Matrix - Ausgang B2
 *
 * Der gesamte Matrix-Puffer besteht aus 4096 Bytes, die in 4 Blcke unterteilt sind:
 *
 * Block 1: 0000-1023 - Pixel   0- 511 (obere Matrix-Hlfte)  - Farbwert 1 (Bit 2-4)  \
 *                      Pixel 512-1023 (untere Matrix-Hlfte) - Farbwert 1 (Bit 5-7)   \ Matrix 0
 * Block 2: 1024-2047 - Pixel   0- 511 (obere Matrix-Hlfte)  - Farbwert 2 (Bit 2-4)   /
 *                      Pixel 512-1023 (untere Matrix-Hlfte) - Farbwert 2 (Bit 5-7)  /
 * Block 3: 2048-3071 - Pixel   0- 511 (obere Matrix-Hlfte)  - Farbwert 1 (Bit 2-4)  \
 *                      Pixel 512-1023 (untere Matrix-Hlfte) - Farbwert 1 (Bit 5-7)   \ Matrix 1
 * Block 4: 3072-4095 - Pixel   0- 511 (obere Matrix-Hlfte)  - Farbwert 2 (Bit 2-4)   /
 *                      Pixel 512-1023 (untere Matrix-Hlfte) - Farbwert 2 (Bit 5-7)  /
 *
 * Es ist immer nur Matrix 0 oder Matrix 1 aktiv. Alle darzustellenden Daten werden in die inaktive Matrix geschrieben, whrend die Interrupt-Routine von Timer 1 und 2 die Daten der aktiven Matrix
 * ausgibt. Somit erfolgt der Bildaufbau im Hintergrund und erst nach Fertigstellung des kompletten Bildes wird durch eine Umschaltung der Matrix der neue Bildinhalt sichtbar.
 *
 * Jedes Byte der Matrix ist gleich aufgebaut und die Daten sind so angeordnet, dass ein Farbbyte direkt mit hoher Geschwindigkeit auf die Matrix ausgegeben werden kann:
 *
 * Bit 0: 0  - nicht genutzt  \ An diesen Ports (PC0 und PC1) befindet sich der TWI-Bus (I2C) des Controllers.
 * Bit 1: 0  - nicht genutzt  / Ich habe die Portbelegung von Adafruit bernommen, obwohl TWI hier nicht verwendet wird.
 * Bit 2: R1 - Rot-Wert von Pixel n (obere Bildhlfte)
 * Bit 3: G1 - Grn-Wert von Pixel n (obere Bildhlfte)
 * Bit 4: B1 - Blau-Wert von Pixel n (obere Bildhlfte)
 * Bit 5: R2 - Rot-Wert von Pixel n + 1024 (untere Bildhlfte)
 * Bit 6: G2 - Grn-Wert von Pixel n + 1024 (untere Bildhlfte)
 * Bit 7: B2 - Blau-Wert von Pixel n + 1024 (untere Bildhlfte)
 *
 * Parameter 1 (r25, r24): Zeiger auf Bitmap
 * Parameter 2 (r22): Multiplex-Zhler
 * Parameter 3 (r20): PWM-Wert
 * Parameter 4 (r18): Matrix-Scan-Modus
 */

#include <avr/io.h>                                     // I/O-Definitionen laden

.global	matrix_out                                      // Definition als globale Funktion

matrix_out:
        in      r21, _SFR_IO_ADDR(PIND)                 // Eingnge von PortD lesen
        andi    r21, _BV(PIND0)                         // Bit-Position 0 filtern (Timing-Jumper)
        sbi     _SFR_IO_ADDR(PORTA), PINA5              // OE auf High (LED-Ausgnge deaktivieren)
        clr     r23                                     // Offset = 0 fr erweiterte Farben (H)
        sbrc    r22, 0                                  // Bit 0 gesetzt? nein -> berspringen
        ldi     r23, 4                                  // sonst Offset 1024 fr erweiterte Farben (H)
        add     r25, r23                                // Offset zum Matrix-Puffer addieren (H)
        lsr     r22                                     // Multiplex-Zhler ein Bit nach rechts, Bits 4-1 enthalten Zeilenadresse
        tst     r18                                     // Matrix-Scan-Modus 1/8 (1)?
        brne    ma1000                                  // ja -> weiter

                                                        // Routine fr Matrix-Scan-Modus 1/16, eine Pixelzeile ausgeben
        andi    r22, 0x0f                               // relevante Bits filtern (4 Bits)
        in      r18, _SFR_IO_ADDR(PORTA)                // PortA lesen
        andi    r18, 0xf0                               // alte Adresse (A-D) lschen
        or      r18, r22                                // neue Adresse aus Multiplex-Zhler einfgen
        out     _SFR_IO_ADDR(PORTA), r18                // auf PortA ausgeben
        ldi     r23, 64                                 // Offset fr eine Pixelzeile
        mul     r22, r23                                // Offset fr auszugebende Pixelzeile berechnen
        add     r24, r0                                 // Zeilenadresse berechnen (L)
        adc     r25, r1                                 // Zeilenadresse berechnen (H)
        movw    r30, r24                                // Ergebnis nach Z kopieren

        ldi     r18, 61                                 // 61 Bytes ausgeben (bis zum LAT-Impuls)
ma0000: ld      r19, z+                                 // Pixelbyte holen
        out     _SFR_IO_ADDR(PORTC), r19                // an PortC ausgeben
        sbi     _SFR_IO_ADDR(PORTB), PINB0              // CLK auf High
        cbi     _SFR_IO_ADDR(PORTB), PINB0              // CLK wieder auf Low
        dec     r18                                     // Bytezhler - 1
        brne    ma0000                                  // alle Bytes bearbeitet? -> Schleife
        tst     r21                                     // Timing-Jumper gesteckt?
        breq    ma0010                                  // ja -> weiter (klassisches Timing)
        sbi     _SFR_IO_ADDR(PORTA), PINA4              // sonst LAT auf High (fr Matrix mit FM6126A-Treiber)

ma0010: ldi     r18, 3                                  // restliche 3 Bytes ausgeben
ma0020: ld      r19, z+                                 // Pixelbyte holen
        out     _SFR_IO_ADDR(PORTC), r19                // an PortC ausgeben
        sbi     _SFR_IO_ADDR(PORTB), PINB0              // CLK auf High
        cbi     _SFR_IO_ADDR(PORTB), PINB0              // CLK wieder auf Low
        dec     r18                                     // Bytezhler - 1
        brne    ma0020                                  // alle Bytes bearbeitet? -> Schleife
        tst     r21                                     // Timing-Jumper gesteckt?
        brne    ma2000                                  // nein -> weiter (neues Timing)
        sbi     _SFR_IO_ADDR(PORTA), PINA4              // sonst LAT auf High (fr Matrix mit Standard-Treiber)
        rjmp    ma2000                                  // Daten bernehmen und Helligkeit steuern

                                                        // Routine fr Matrix-Scan-Modus 1/8, zwei Pixelzeilen ausgeben, n+8 und n
ma1000: andi    r22, 0x07                               // relevante Bits filtern (3 Bits)
        in      r18, _SFR_IO_ADDR(PORTA)                // PortA lesen
        andi    r18, 0xf8                               // alte Adresse (A-C) lschen
        or      r18, r22                                // neue Adresse aus Multiplex-Zhler einfgen
        out     _SFR_IO_ADDR(PORTA), r18                // auf PortA ausgeben
        ldi     r23, 64                                 // Offset fr eine Pixelzeile
        mul     r22, r23                                // Offset fr auszugebende Pixelzeile berechnen
        add     r24, r0                                 // Zeilenadresse berechnen (L)
        adc     r25, r1                                 // Zeilenadresse berechnen (H)
        movw    r30, r24                                // Ergebnis nach Z kopieren
        subi    r31, -2                                 // Offset 512 (+ 8 Zeilen)
        ldi     r18, 64                                 // 64 Bytes ausgeben (Zeile n+8)
ma1010: ld      r19, z+                                 // Pixelbyte holen
        out     _SFR_IO_ADDR(PORTC), r19                // an PortC ausgeben
        sbi     _SFR_IO_ADDR(PORTB), PINB0              // CLK auf High
        cbi     _SFR_IO_ADDR(PORTB), PINB0              // CLK wieder auf Low
        dec     r18                                     // Bytezhler - 1
        brne    ma1010                                  // alle Bytes bearbeitet? -> Schleife

        movw    r30, r24                                // originale Zeilenadresse nach Z kopieren
        ldi     r18, 61                                 // 61 Bytes ausgeben (Zeile n, bis zum LAT-Impuls)
ma1020: ld      r19, z+                                 // Pixelbyte holen
        out     _SFR_IO_ADDR(PORTC), r19                // an PortC ausgeben
        sbi     _SFR_IO_ADDR(PORTB), PINB0              // CLK auf High
        cbi     _SFR_IO_ADDR(PORTB), PINB0              // CLK wieder auf Low
        dec     r18                                     // Bytezhler - 1
        brne    ma1020                                  // alle Bytes bearbeitet? -> Schleife
        tst     r21                                     // Timing-Jumper gesteckt?
        breq    ma0010                                  // ja -> weiter (klassisches Timing)
        sbi     _SFR_IO_ADDR(PORTA), PINA4              // sonst LAT auf High (fr Matrix mit FM6126A-Treiber)

ma1030: ldi     r18, 3                                  // restliche 3 Bytes ausgeben
ma1040: ld      r19, z+                                 // Pixelbyte holen
        out     _SFR_IO_ADDR(PORTC), r19                // an PortC ausgeben
        sbi     _SFR_IO_ADDR(PORTB), PINB0              // CLK auf High
        cbi     _SFR_IO_ADDR(PORTB), PINB0              // CLK wieder auf Low
        dec     r18                                     // Bytezhler - 1
        brne    ma0020                                  // alle Bytes bearbeitet? -> Schleife
        tst     r21                                     // Timing-Jumper gesteckt?
        brne    ma2000                                  // nein -> weiter (neues Timing)
        sbi     _SFR_IO_ADDR(PORTA), PINA4              // sonst LAT auf High (fr Matrix mit Standard-Treiber)
        rjmp    ma2000                                  // Daten bernehmen und Helligkeit steuern

ma2000: cbi     _SFR_IO_ADDR(PORTA), PINA4              // LAT auf Low
        cbi     _SFR_IO_ADDR(PORTA), PINA5              // OE auf Low (LED-Ausgnge aktivieren)
        cpi     r20, 3                                  // PWM-Wert >= 3?
        brcc    ma2050                                  // ja -> weiter verzweigen
        cpi     r20, 0                                  // PWM-Wert = 0?
        brne    ma2040                                  // nein -> weiter verzweigen
ma2010: nop                                             // sonst Wartezeit 7 Taktzyklen
ma2020: nop
ma2030: sbi     _SFR_IO_ADDR(PORTA), PINA5              // OE auf High (LED-Ausgnge deaktivieren)
        ret

ma2040: cpi     r20, 1                                  // PWM-Wert = 1?
        brne    ma2030                                  // nein -> Wartezeit 9 Taktzyklen, sonst 8 Taktzyklen
        sbi     _SFR_IO_ADDR(PORTA), PINA5              // OE auf High (LED-Ausgnge deaktivieren)
        ret

ma2050: cpi     r20, 6                                  // PWM-Wert >= 6?
        brcc    ma2070                                  // ja -> weiter verzweigen
        cpi     r20, 3                                  // PWM-Wert = 3?
        brne    ma2060                                  // nein -> weiter verzweigen
        rjmp    ma2020                                  // sonst Wartezeit 11 Taktzyklen
ma2060: cpi     r20, 4                                  // PWM-Wert = 4?
        breq    ma2020                                  // ja -> Wartezeit 13 Taktzyklen
        rjmp    ma2010                                  // sonst Wartezeit 15 Taktzyklen
ma2070: cpi     r20, 21                                 // PWM-Wert >= 21?
        brcc    ma2250                                  // ja -> weiter verzweigen
        cpi     r20, 6                                  // PWM-Wert = 6?
        brne    ma2080                                  // nein -> weiter verzweigen
        ldi     r19, 1                                  // sonst Wartezeit 18 Taktzyklen
        rjmp    ma2220                                  // weiter zur Warteschleife
ma2080: cpi     r20, 7                                  // PWM-Wert = 7?
        brne    ma2090                                  // nein -> weiter verzweigen
        ldi     r19, 1                                  // sonst Wartezeit 21 Taktzyklen
        rjmp    ma2220                                  // weiter zur Warteschleife
ma2090: cpi     r20, 8                                  // PWM-Wert = 8?
        brne    ma2100                                  // nein -> weiter verzweigen
        ldi     r19, 1                                  // sonst Wartezeit 24 Taktzyklen
        rjmp    ma2220                                  // weiter zur Warteschleife
ma2100: cpi     r20, 9                                  // PWM-Wert = 9?
        brne    ma2110                                  // nein -> weiter verzweigen
        ldi     r19, 2                                  // sonst Wartezeit 28 Taktzyklen
        rjmp    ma2240                                  // weiter zur Warteschleife
ma2110: cpi     r20, 10                                 // PWM-Wert = 10?
        brne    ma2120                                  // nein -> weiter verzweigen
        ldi     r19, 2                                  // sonst Wartezeit 32 Taktzyklen
        rjmp    ma2230                                  // weiter zur Warteschleife
ma2120: cpi     r20, 11                                 // PWM-Wert = 11?
        brne    ma2130                                  // nein -> weiter verzweigen
        ldi     r19, 2                                  // sonst Wartezeit 36 Taktzyklen
        rjmp    ma2220                                  // weiter zur Warteschleife
ma2130: cpi     r20, 12                                 // PWM-Wert = 12?
        brne    ma2140                                  // nein -> weiter verzweigen
        ldi     r19, 3                                  // sonst Wartezeit 41 Taktzyklen
        rjmp    ma2230                                  // weiter zur Warteschleife
ma2140: cpi     r20, 13                                 // PWM-Wert = 13?
        brne    ma2150                                  // nein -> weiter verzweigen
        ldi     r19, 4                                  // sonst Wartezeit 46 Taktzyklen
        rjmp    ma2240                                  // weiter zur Warteschleife
ma2150: cpi     r20, 14                                 // PWM-Wert = 14?
        brne    ma2160                                  // nein -> weiter verzweigen
        ldi     r19, 4                                  // sonst Wartezeit 51 Taktzyklen
        rjmp    ma2220                                  // weiter zur Warteschleife
ma2160: cpi     r20, 15                                 // PWM-Wert = 15?
        brne    ma2170                                  // nein -> weiter verzweigen
        ldi     r19, 5                                  // sonst Wartezeit 57 Taktzyklen
        rjmp    ma2220                                  // weiter zur Warteschleife
ma2170: cpi     r20, 16                                 // PWM-Wert = 16?
        brne    ma2180                                  // nein -> weiter verzweigen
        ldi     r19, 6                                  // sonst Wartezeit 63 Taktzyklen
        rjmp    ma2220                                  // weiter zur Warteschleife
ma2180: cpi     r20, 17                                 // PWM-Wert = 17?
        brne    ma2190                                  // nein -> weiter verzweigen
        ldi     r19, 7                                  // sonst Wartezeit 69 Taktzyklen
        rjmp    ma2220                                  // weiter zur Warteschleife
ma2190: cpi     r20, 18                                 // PWM-Wert = 18?
        brne    ma2200                                  // nein -> weiter verzweigen
        ldi     r19, 9                                  // sonst Wartezeit 76 Taktzyklen
        rjmp    ma2240                                  // weiter zur Warteschleife
ma2200: cpi     r20, 19                                 // PWM-Wert = 19?
        brne    ma2210                                  // nein -> weiter zu PWM-Wert 20
        ldi     r19, 10                                 // sonst Wartezeit 84 Taktzyklen
        rjmp    ma2220                                  // weiter zur Warteschleife
ma2210: ldi     r19, 13                                 // sonst Wartezeit 92 Taktzyklen
ma2220: nop                                             // Warteschleife, zustzlicher Taktzyklus
ma2230: nop                                             // Warteschleife, zustzlicher Taktzyklus
ma2240: dec     r19                                     // Zhler vermindern
        brne    ma2240                                  // Zhler abgelaufen? nein -> weitere Schleife (weitere 3 Taktzyklen)
        sbi     _SFR_IO_ADDR(PORTA), PINA5              // OE auf High (LED-Ausgnge deaktivieren)
        ret
                                                        // ab hier wird ein Timer fr die Abschaltung der LED-Ausgnge verwendet
ma2250: cpi     r20, 31                                 // PWM-Wert >= 31
        brcc    ma2270                                  // ja -> weiter aufbereiten
        subi    r20, 21                                 // sonst PWM-Wert 21-30 in Wertebereich 0-9 umwandeln, entspricht einem Raster von 8 Taktzyklen
        ldi     r19, 0xf8                               // Timer-Basiswert laden, entspricht 64 Taktzyklen, eine Verkleinerung des Wertes um 1 erhht die Zeit um 8 Taktzyklen
        sub     r19, r20                                // umgewandelten PWM-Wert vom Basiswert abziehen
        ldi     r20, 1 << CS21                          // Timer2 auf Vorteiler 8 setzen
        nop                                             // Ausgleich fr Wartezeit von 100 Taktzyklen bei PWM-Wert 21
ma2260: sts     _SFR_IO_ADDR(TCCR2B + 32), r20          // Timer2 Vorteiler setzen
        ldi     r20, 1 << PSRASY
        out     _SFR_IO_ADDR(GTCCR), r20                // Timer2 Vorteiler zurcksetzen
        sts     _SFR_IO_ADDR(TCNT2 + 32), r19           // Timer2 Zhler setzen
        ldi     r20, 1 << TOIE2
        sts     _SFR_IO_ADDR(TIMSK2 + 32), r20          // Timer2 Overflow Interrupt aktivieren
        ldi     r20, 1 << TOV2
        out     _SFR_IO_ADDR(TIFR2), r20                // Timer2 Overflow Interrupt Flag lschen
        ret

ma2270: cpi     r20, 42                                 // PWM-Wert >= 42
        brcc    ma2280                                  // ja -> weiter aufbereiten
        subi    r20, 31                                 // sonst PWM-Wert 31-41 in Wertebereich 0-10 umwandeln
        add     r20, r20                                // Wert verdoppeln, entspricht einem Raster von 16 Taktzyklen
        ldi     r19, 0xee                               // Timer-Basiswert laden, entspricht 144 Taktzyklen, eine Verkleinerung des Wertes um 1 erhht die Zeit um 8 Taktzyklen
        sub     r19, r20                                // umgewandelten PWM-Wert vom Basiswert abziehen
        ldi     r20, 1 << CS21                          // Timer2 auf Vorteiler 8 setzen
        rjmp    ma2260                                  // Timer2 starten

ma2280: subi    r20, 42                                 // sonst PWM-Wert 42-105 in Wertebereich 0-63 umwandeln
        ldi     r19, 0xf6                               // Timer-Basiswert laden, entspricht 320 Taktzyklen, eine Verkleinerung des Wertes erhht die Zeit um 32 Taktzyklen
        sub     r19, r20                                // umgewandelten PWM-Wert vom Basiswert abziehen
        ldi     r20, (1 << CS21) | (1 << CS20)          // Timer2 auf Vorteiler 32 setzen
        rjmp    ma2260                                  // Timer2 starten
